
%{

exception Yyerr of string * string;
fun yyerr (yyd, msg, s) = (hlex_begin (yyd, 0);
                           hlex_flush (yyd, hlex_current_buffer yyd);
                           raise Yyerr (msg, s));

fun read_octal (s, i, n) =
    let val ir = ref i
        val acc = ref 0
    in
       while !ir<n do
         (acc := lsl (!acc, 3) + ordof (s, !ir)-48;
          inc ir);
       !acc
    end;

val yyvalue = ref "";
%}

%x str

%%

    val string_buf = ref (outstring "");

\"      #seek (!string_buf) 0; #truncate (!string_buf) (); hlex_begin (yyd, str); continue

<str>\"        { (* saw closing quote - all done *)
            hlex_begin (yyd, INITIAL);
            (* return string constant token type and
             * value to parser
             *)
            yyvalue := #convert (!string_buf) ();
            return 1
            }

<str>\n        {
            (* error - unterminated string constant *)
            (* generate error message *)
            yyerr (yyd, "unterminated string constant", "")
            }

<str>\\[0-7]{1,3} { (* octal escape sequence *)
                   let val yytext = hlex_text yyd
                       val result = read_octal (yytext, 1, size yytext)
		   in
		      if result > 0xff
		         then yyerr(yyd, "constant out of bounds: ", yytext)
		      else #put (!string_buf) (chr result)
		   end; continue
            }

<str>\\[0-9]+ {
            (* generate error - bad escape sequence; something
             * like '\48' or '\0777777'
             *)
            yyerr (yyd, "bad escape sequence", hlex_text yyd)
            }

<str>\\n  #put (!string_buf) "\n"; continue
<str>\\t  #put (!string_buf) "\t"; continue
<str>\\r  #put (!string_buf) "\^M"; continue
<str>\\b  #put (!string_buf) "\^H"; continue
<str>\\f  #put (!string_buf) "\f"; continue

<str>\\(.|\n)  #put (!string_buf) (substr (hlex_text yyd, 1, hlex_length yyd)); continue

<str>[^\\\n\"]+  #put (!string_buf) (hlex_text yyd); continue
