%{
open "kw_h";

val string_buf = ref (outstring "");
val old_mode = ref 0;
val yylloc = ref |[first_line=0, last_line=0, first_column=0, last_column=0]|;
val yyfilename = ref "";

(* fun yywrap () = true; (* stop *) *)

exception Yyerr of string * string;
fun yyerr (yyd, msg, s) = (hlex_begin (yyd, 0);
                           hlex_flush (yyd, hlex_current_buffer yyd);
                           raise Yyerr (msg, s));

fun read_octal (s, i, n) =
    let val ir = ref i
        val acc = ref 0
    in
       while !ir<n do
         (acc := lsl (!acc, 3) + ordof (s, !ir)-48;
          inc ir);
       !acc
    end;

fun read_dec (s, i, n) =
    if i>=n
       then 0
    else 10*(ordof (s, i)-48)+read_dec (s, i+1, n);

%}

%x COMMENT
%x STRING
%x LINE LINE2

SPACE [\ \t\f]
DIG [0-9]
INT ({DIG}+)
FMT [\ \t\n\f]
LETTER [A-Za-z]
ALPHA [A-Za-z0-9\_]
SYMB [\!\%\&\$\#\+\-\/\:\<\=\>\?\@\\\~\'\^\|\*\`]
ID (\_*{LETTER}{ALPHA}*)

%%

symk   return kw_symk
asymk1 return kw_asymk1
asymk2 return kw_asymk2
sk     return kw_sk
pubk   return kw_pubk
privk  return kw_privk
nudge  return kw_nudge

write  return kw_write
read   return kw_read
let    return kw_let
new    return kw_new
sym    return kw_sym
key    return kw_key
asym   return kw_asym
pair   return kw_pair
role   return kw_role

secret return kw_secret

simulate return kw_simulate
multi return kw_multi

reset return kw_reset
break return kw_break

msg    return kw_msg
K      return kw_K
D      return kw_D
only   return kw_only

"{"	return kw_open_brace
"}"	return kw_close_brace
"["	return kw_open_bracket
"]"	return kw_close_bracket
"("	return kw_open_paren
")"	return kw_close_paren
"|"     return kw_pipe
"<"	return kw_open_angle
">"	return kw_close_angle
","	return kw_comma
":"     return kw_colon
";"     return kw_semi
"="     return kw_equal
"/"     return kw_slash

{ID}	return kw_identifier

\n+	continue
{SPACE}+ continue

<INITIAL,LINE2>\"	(#seek (!string_buf) 0; #truncate (!string_buf) ();
			       old_mode := hlex_start yyd; hlex_begin (yyd, STRING); continue
			       )
<STRING>{
  \"	(hlex_begin (yyd, !old_mode);
	 if !old_mode=INITIAL
            then return kw_string_const
         else continue)
  \n+	yyerr (yyd, "unterminated string constant", "")
  \\[0-7]{1,3}	( (* octal escape sequence *)
                   let val yytext = hlex_text yyd
                       val result = read_octal (yytext, 1, size yytext)
		   in
		      if result > 0xff
		         then yyerr(yyd, "constant out of bounds: ", yytext)
		      else #put (!string_buf) (chr result)
		   end; continue
                   )
  \\[0-9]+	yyerr(yyd, "bad escape sequence: ", hlex_text yyd)
  \\n	#put (!string_buf) "\n"; continue
  \\t	#put (!string_buf) "\t"; continue
  \\r	#put (!string_buf) "\^M"; continue
  \\b	#put (!string_buf) "\^H"; continue
  \\f	#put (!string_buf) "\f"; continue
  \\(.|\n)	#put (!string_buf) (chr (ordof (hlex_text yyd, 1))); continue
  [^\\\n\"]+	#put (!string_buf) (hlex_text yyd); continue
  <<EOF>>	yyerr(yyd, "unterminated string","")
}

"//"[^\n]*\n+	continue
"/*"	hlex_begin (yyd, COMMENT); continue
<COMMENT>{
  [^*\n]*	continue (* eat anything that's not a '*' *)
  "*"+[^*/\n]*  continue (* eat up '*'s not followed by '/'s *)
  \n+           continue
  "*"+"/"       hlex_begin (yyd, INITIAL); continue
  <<EOF>>	yyerr(yyd, "unterminated comment","")
}

^\.	return kw_ready
^\#	#seek (!string_buf) 0; #truncate (!string_buf) (); hlex_begin (yyd, LINE); continue
<LINE>{
  {INT}	( let val yytext = hlex_text yyd
              val n = read_dec (yytext, 0, size yytext)
          in
             yylloc := |[ first_line = n, last_line = n, first_column = 0, last_column = 0]|
          end;
          hlex_begin (yyd, LINE2); continue)
  \n+	|
  <<EOF>>	hlex_begin (yyd, INITIAL); continue
  .	continue
}

<LINE2>{
  \n+	|
  <<EOF>>	(let val s = #convert (!string_buf) ()
                 in
                    if s=""
                       then ()
                    else yyfilename := s
                 end;
                 hlex_begin (yyd, INITIAL); continue)
  .	continue
}

<*>.	yyerr (yyd, "unrecognized lone character ('" ^ hlex_text yyd ^ "')", "")
<<EOF>>	return kw_eof

%%
